/////////////////////////////////////////////////////////////////////////////
//Es wird ein Poincaré-Schnitt des Doppelpendel ohne Reibung gezeichnet. 
//Näherung: Runge-Kutta 4 
//Genügend Rechenzeit vorausgesetzt, sind wegen der Verwendung von double 
//und sehr kleinen Zeitschritten die Lösungen auch für kleinste Zeitschritte sehr genau .
//
//In dieser Version werden (AnzahlCore-1) Threads (AnzahlCore/4) mal gestartet
//Jeder Thread berechnet für vier Startpunkte M/4 Folgepunkte
//Beschleunigt wird der Vorgang neben der Multithread-Technik auch durch 
//mathematisch nicht ganz korrekten zwei Zeitschritte. Der kleinere Zeitschritt 
//wird nur beim Durchgang durch PI beim großen Zeitschritt verwendet.
//Der letzte Zustand vor dem Durchgang ist der Start der kleinen Zeitschritte
//Vorteil: Zehnmal so schnell. Nachteil: Durch Wechsel der Zeitschritte ist 
//die Lösung möglicherweise weniger exakt.
//
//Beispiel : Bei 8 Core werden bei jedem Bild 7*8/4*4 = 56 Startpunkte erzeugt. 
//Mehr als (Anzahl Core) Neustarts der Threads ist nicht sinnvoll, weil 
//der Process sonst sehr langsam wird. Bisher kein Lösung für dieses Problem(chen) gefunden!
//Klingt kompliziert, ist aber einfach: Dadurch, dass man nur (AnzahlCore/4) mal die 
//Threads startet, kann man noch drei mal zoomen, ohne, dass damit über die Hürde (AnzahlCore) 
//Starts zu kommen. Der einzelne Thread hat nun natürlich vier mal mehr zu tun - 
//dennoch aber gibt es keine Verlangsamung der Rechnung wie nach dem S-ten 
//Neustart aller Threads.
//
//Hier wurde M = 2000*S gewählt (M/4 ist Anzahl der Punkte pro Thread)
//Für langsamer Rechner sollte man M = 100*S wählen
////////////////////////////////////////////////////////////////////////////////////////

// Attribute:

double w1,w2; //Winkel von der Senkrechten weg
double W1,W2; //Für Runge-Kutta verwendet: Winkel in Ruhelage: PI
double g1,g2; //Winkelgeschwindigkeiten    
double dt1 = 0.001;      // Zeitschritt1
double dt2 = 0.0001;     // Zeitschritt2
double dt;
double zwW, zwG, Wi, Ge;
double WiNeu,GeNeu,zwWNeu, zwGNeu;
double masse;
double E; //Gesamtenergie mit Nullniveau in Ruhelage
double L = 180;
double G = 9.81;         // Ortfaktor für Schwerkraft
double FensterDim = 1500;
double ruhelage = FensterDim/2;         // Ruhelag in x-Koordinate
double EKontrolleA;//Gesamtenergie 
double wK,gK; 
color farbe; double f =0;
double D1,D2; //Diskriminaten für die Berechnung der fehlenden Winkelgeschwindigkeit bei gleicher Energie
int M, S,K, s; //M: Anzahl Punkte pro Startpunkt, S: Anzahl Startpunkte / Anzahl Threads, Zähler für S
int P; //nach P Folgepunkten neuen Startpunkt wählen
boolean finished;
ArrayList<PVector> punkteW = new ArrayList<PVector>(); //Zum speichern der Kurvenpunkte
ArrayList<MyThread> workers = new ArrayList<MyThread>();
  
///////////////////////////////////////////////////////////////////////////

void setup() {
  size(1500,1500);
  background(0);  
  colorMode(HSB,256);
  int processors = Runtime.getRuntime().availableProcessors();
  S = processors; //Anzahl der Core
  s = 0;
  E = 1800; //Gesamtenergie
  M = 2000*S;//Anzahl Punkte für einen Thread ; je Thread vier Startpunkte; M abhängig von der Anzahl der Core
  dt = dt1;
  //Wenn man K = S/2 wählt, kann man einmal mit gleicher Geschwindigkeit zoomen
  K = S/4; //K: Anzahl der Wiederholungen. K>S wird langsam...  
  println("Der Prozessor hat " + S + " Kerne"); 
  println(S-1 + " Threads werden "+ K  +" mal gestartet"); 
  //Hiermit stellt man den gewünschten Bereich ein(hier Gesamtbild):
  Wi  = PI; //Winkel-Zentrum
  zwW = PI/3+0.1;//Map-Bereich für Delta-Winkel oberes Pendel
  Ge  = -0.05;//WGeschw Zentrum
  zwG = 0.32;//Map-Bereich für WGeschw.
  threadsStarten();  
}
/////////////////////////////////////////////////////////////////
void threadsStarten(){
 while(s<K){ //K mal starten ...
      int nummer = s+1;
      println("Nummer: "+ nummer);
      for (int i=0; i<S-1; i++) { //S-1 Threads
        MyThread t = new MyThread(E,M);
        // nun starte den Thread:
        t.start(); 
        workers.add(t); 
       }
             
       do  {
        // prüfe, ob alle Prozesse schon fertig sind
        finished = true;
        for (MyThread worker : workers) 
            if (worker.isAlive()) finished = false;  
              
         // warte ein wenig, damit nicht zu viele Nachfragen
         // den Rechner blockieren
        delay(10);
      } while (!finished);     
      
       Thread.currentThread().interrupt();      
       for (MyThread worker : workers){
          punkteW = worker.punkte;      
          punkteZeichnen();   
          
       }       
       s++; 
  }//end of while 
 
  workers.clear();//workers löschen und danach neu definieren
  
}
/////////////////////////////////////////////////////////
void draw() { 
 
  
}
//////////////////////////////////////////////////////////////////////
void punkteZeichnen(){  
  if(punkteW.size() !=0){
    f = random(256); 
    farbe = color((int)f,256,256);
    stroke(farbe);
    strokeWeight(1);
    for(int i =0; i< punkteW.size()-1 ;i++){    
        if(i==punkteW.size()/4){
          f = random(256);
          farbe = color((int)f,256,256);
          stroke(farbe);;       
        }
        point(punkteW.get(i).x,punkteW.get(i).y);       
     }  
    punkteW.clear();
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////
void mousePressed() { 
  zwWNeu = zwW/5;//Map-Bereich für Delta-Winkel oberes Pendel
  zwGNeu = zwG/5;//Map-Bereich für WGeschw.
  finished = false;
   background(0);  
  WiNeu=    map(mouseX,(float)0,(float)FensterDim,(float)(Wi-zwW),(float)(Wi+zwW));; //Winkel-Zentrum  
  GeNeu  =  map(mouseY, (float)FensterDim,(float)0,(float)(Ge-zwG),(float)(Ge +zwG));//WGeschw Zentrum
  //Zur Kontrolle
  println(Wi,Ge,zwW,zwG);
  println(WiNeu,GeNeu,zwWNeu, zwGNeu);
  s = 0;
  Wi  = WiNeu; //Winkel-Zentrum
  zwW = zwWNeu;//Map-Bereich für Delta-Winkel oberes Pendel
  Ge  = GeNeu;//WGeschw Zentrum
  zwG = zwGNeu;//Map-Bereich für WGeschw.
  threadsStarten();
}

////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
public class MyThread extends Thread {
  int n = 4;            // Anzahl der Variablen
  double E;
  int M;
  int P,p;
  int m; 
  double w1,w2,W1,W2; 
  double g1,g2;
  double[] ZVaktuell ; // Initialisierung w1,w2,g1,g2
  ArrayList<PVector> punkte ;
  boolean fertig,neuerSP;
  double wVor0,wVor1,wNach1,gVor0,gVor1;
 // int anzahlSchritte =0;
  // Thread anlegen und Parameter übergeben
  public MyThread(double E, int M){
     this.E = E;
     this.M = M;    
     fertig = false;
     neuerSP= false;
     masse =1;
     m = 0;
     p = 0;
     ZVaktuell = new double[4];
     punkte = new ArrayList<PVector>();
  }
/////////////////////////////////////////////////////////////
  public void run(){
    neuStarten(); 
    while(!fertig){
       if(neuerSP) {         
         neuStarten();
         neuerSP = false;
       }
       folgePunkteBerechnen();      
    } 
 }
//////////////////////////////////////////////////////////
double[] ZVaktualisieren(double[] Y) {
    double[] ZustandsVektor = new double[n]; 
    ZustandsVektor[0]=Y[2];
    ZustandsVektor[1]=Y[3];
    // benötigte Werte :
    double sinDiff = Math.sin(Y[0]-Y[1]);
    double cosDiff = Math.cos(Y[0]-Y[1]);
    double cosProdukt = cosDiff * cosDiff;
    double s0 = Math.sin(Y[0]);
    double s1 = Math.sin(Y[1]);
    double alpha = - Y[3]*Y[3]*sinDiff + 2*G/L*s0 ;
    double beta = Y[2]*Y[2]*sinDiff + G/L*s1 ;
    ZustandsVektor[2]= 1 / (2-cosProdukt) * (alpha - cosDiff * beta); //Geschwindigkeit 1
    ZustandsVektor[3]= 1 / (2-cosProdukt) * (2*beta - cosDiff * alpha);//Geschwindigkeit 2 
    return ZustandsVektor;
}
////////////////////////////////////////////////////////////////////////////
void pendelUpdate(){
    // Zustand nach nächstem Zeitschritt mit Runge-Kutta 4 berechnen
    // "double[] ZVaktualisieren" Methode der zeitlichen Ableitung dy/dt (4-dimensional). 
    // Die DGL hängt hier nicht explizit von t ab
    // anzahlSchritte++;
    //Die folgende Zeile ist Bestandteil für das Zeichnen eines Poincaré-Schnitts   
     wVor0 = ZVaktuell[0];
     wVor1 = ZVaktuell[1];
     gVor0 = ZVaktuell[2];
     gVor1 = ZVaktuell[3];
     RK();
     wNach1 = ZVaktuell[1];
     if(wVor1 < PI && wNach1 > PI)    {
       if(dt != dt2) {
         dt = dt2;
         //Zurück zum ersten Wert und von dort in kleineren Schritten weiter
         ZVaktuell[0] = wVor0;
         ZVaktuell[1] = wVor1;
         ZVaktuell[2] = gVor0;
         ZVaktuell[3] = gVor1;
         
       }else {
          wK = map((float)ZVaktuell[0],(float)(Wi-zwW),(float)(Wi+zwW),(float)0,(float)FensterDim);  //Koordinaten der Poincaré-Schnitt Punkte 
          gK = map((float)ZVaktuell[2],(float)(Ge-zwG),(float)(Ge +zwG),(float)FensterDim,(float)0);   
          punkte.add(new PVector((float)wK,(float)gK));
          m++;  
          dt=dt1; 
       }   
     }
    //Winkel immer zwischen 0 und 2*PI 
    if(ZVaktuell[0] > 2*PI) ZVaktuell[0] = ZVaktuell[0] - 2*PI;
    if(ZVaktuell[0] < 0)    ZVaktuell[0] = ZVaktuell[0] + 2*PI;
    if(ZVaktuell[1] > 2*PI) ZVaktuell[1] = ZVaktuell[1] - 2*PI;
    if(ZVaktuell[1] < 0)    ZVaktuell[1] = ZVaktuell[1] + 2*PI;
   
    W1 = ZVaktuell[0]; w1 = PI-W1;
    W2 = ZVaktuell[1]; w2 = PI-W2;
    g1 = ZVaktuell[2];
    g2 = ZVaktuell[3]; 
 }
///////////////////////////////////////////////////////////////////////////////
//Nach der Hälfte der zu errechnenden Punkte wird ein neuer Startpunkt gesetzt 
  void folgePunkteBerechnen(){ 
    pendelUpdate();
    if(m == M/4){
      neuerSP = true;m++;
    }
    if(m >= M){
      fertig = true; 
    }   
   }
////////////////////////////////////////////////////////////////////////////
void RK(){
  double[] K1 = ZVaktualisieren(ZVaktuell); 
    // K2 =ZustandsVektor(t0 + dt/2, ZVaktuell + dt/2 * K1)
    double[] ZVtemp = new double[n];
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K1[i]*dt*0.5;    
    }
    double[] K2 = ZVaktualisieren(ZVtemp);
    // K3 =ZustandsVektor(t0 + dt/2, ZVaktuell + dt/2 * K2)
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K2[i]*dt*0.5;
    }
    double[] K3 = ZVaktualisieren(ZVtemp);
    // K4 =ZustandsVektor(t0 + dt, ZVaktuell + dt * K3)
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K3[i]*dt;
    }
    double[] K4 = ZVaktualisieren(ZVtemp);
    // Zusammenfassung
    //  ZVaktuell = ZVaktuell + dt/6 * (K1 + 2*(K2+K3) + K4)
    for (int i=0; i<n; i++) {
      ZVaktuell[i] = ZVaktuell[i] + dt / 6.0 * (K1[i] + 2.0*(K2[i] + K3[i]) + K4[i]);
    }  
}

///////////////////////////////////////////////////////////////////////////////
  void neuStarten(){
    dt = dt1;//Mit größerem Zeitabschnitt starten
    zufallsWerteBestimmen(1);
    if(g2 == -1000) zufallsWerteBestimmen(3);     
    ZVaktuell[0] = W1;
    ZVaktuell[1] = W2;         
    ZVaktuell[2] = g1;
    ZVaktuell[3] = g2; 
  }  
////////////////////////////////////////////////////////////////
void zufallsWerteBestimmen(float nenner){
    w1=random((float)(Wi+PI-zwW/nenner),(float)(Wi+PI+zwW/nenner));    
    W1 = w1-PI;
    //Zweites Pendel befindet sich in der Nulllage, um im Phasendiagramm im Zoom-Ausschnitt zu sein
    W2=PI;
    g1=random((float)(Ge-zwG/nenner),(float)(Ge+zwG/nenner)); 
    g2 = berechneG2();    
}
////////////////////////////////////////////////////////////////////
  double berechneG2(){//Damit Energie konstant ist   
     D2 = g1*g1*Math.cos(w2-w1)*Math.cos(w2-w1) -2*g1*g1+ 2*G*(2*Math.cos(w1)+Math.cos(w2)-3)/L + 2*E/(L*L*masse);
     if(D2 >=0)  
        g2 = -g1*Math.cos(w1-w2) + Math.sqrt(D2); 
     else g2 = -1000;// Wert kommt sonst nicht vor 
    return g2; 
  }
///////////////////////////////////////////////////////////////////////////////
}
